# Category metadata.

# Category icon show in the menu
ICON = "icons/Category-Fusion.svg"

# Background color for category background in menu
# and widget icon background in workflow.
BACKGROUND = "light-green"
